use std::collections::HashSet;

use rayon::iter::ParallelBridge;
use rayon::prelude::*;

#[derive(Debug, PartialEq, Eq, Clone, Copy, Hash)]
enum Direction {
    Up,
    Right,
    Down,
    Left,
}
impl Direction {
    fn next(self) -> Self {
        match self {
            Direction::Up => Direction::Right,
            Direction::Right => Direction::Down,
            Direction::Down => Direction::Left,
            Direction::Left => Direction::Up,
        }
    }
}
#[derive(Debug, PartialEq, Eq, Clone, Copy)]
enum Field {
    Empty,
    Blocked,
    Guard(Direction),
}
fn parse_field(c: char) -> Field {
    match c {
        '.' => Field::Empty,
        '^' => Field::Guard(Direction::Up),
        '>' => Field::Guard(Direction::Right),
        'v' => Field::Guard(Direction::Down),
        '<' => Field::Guard(Direction::Left),
        '#' => Field::Blocked,
        c => panic!("Unknown field: {c}"),
    }
}

pub fn pt1(input: &str) -> usize {
    let mut visited_fields: HashSet<(usize, usize)> = HashSet::new();
    let mut current_pos = (0, 0);
    let mut current_direction = Direction::Up;
    let grid: Vec<_> = parse_grid(input, &mut current_direction, &mut current_pos);

    loop {
        tracing::debug!("{current_pos:?} [{current_direction:?}]");
        let next_pos: (usize, usize) = match current_direction {
            Direction::Up => {
                if current_pos.1 == 0 {
                    visited_fields.insert(current_pos);
                    break;
                }
                (current_pos.0, current_pos.1 - 1)
            }
            Direction::Right => {
                if grid[current_pos.1].len() - 1 == current_pos.0 {
                    visited_fields.insert(current_pos);
                    break;
                }
                (current_pos.0 + 1, current_pos.1)
            }
            Direction::Down => {
                if grid.len() - 1 == current_pos.1 {
                    visited_fields.insert(current_pos);
                    break;
                }
                (current_pos.0, current_pos.1 + 1)
            }
            Direction::Left => {
                if current_pos.0 == 0 {
                    visited_fields.insert(current_pos);
                    break;
                }
                (current_pos.0 - 1, current_pos.1)
            }
        };
        match grid
            .get(next_pos.1)
            .and_then(|l| l.get(next_pos.0))
            .unwrap()
        {
            Field::Empty => {
                visited_fields.insert(current_pos);
                current_pos = next_pos;
                continue;
            }
            Field::Blocked => {
                current_direction = current_direction.next();
                continue;
            }
            Field::Guard(_) => unreachable!(),
        }
    }
    visited_fields.len()
}

pub fn pt2(input: &str) -> usize {
    let mut initial_position = (0, 0);
    let mut initial_direction = Direction::Up;
    let original_grid: Vec<_> = parse_grid(input, &mut initial_direction, &mut initial_position);

    original_grid
        .iter()
        .enumerate()
        .flat_map(|(y, row)| (0..row.len()).map(move |x| (x, y)))
        .par_bridge()
        .filter(|&(x, y)| {
            if (x, y) == initial_position {
                return false;
            }
            let mut grid = original_grid.clone();
            grid[y][x] = Field::Blocked;
            let mut current_pos = initial_position;
            let mut current_direction = initial_direction;
            let mut visited_fields: HashSet<((usize, usize), Direction)> = HashSet::new();
            let mut is_valid = true;
            'current_attempt: loop {
                tracing::debug!("{current_pos:?} [{current_direction:?}]");
                if !visited_fields.insert((current_pos, current_direction)) {
                    is_valid = false;
                    break 'current_attempt;
                }
                let next_pos = match current_direction {
                    Direction::Up => {
                        if current_pos.1 == 0 {
                            break 'current_attempt;
                        }
                        (current_pos.0, current_pos.1 - 1)
                    }
                    Direction::Right => {
                        if grid[current_pos.1].len() - 1 == current_pos.0 {
                            break 'current_attempt;
                        }
                        (current_pos.0 + 1, current_pos.1)
                    }
                    Direction::Down => {
                        if grid.len() - 1 == current_pos.1 {
                            break 'current_attempt;
                        }
                        (current_pos.0, current_pos.1 + 1)
                    }
                    Direction::Left => {
                        if current_pos.0 == 0 {
                            break 'current_attempt;
                        }
                        (current_pos.0 - 1, current_pos.1)
                    }
                };
                match grid
                    .get(next_pos.1)
                    .and_then(|l| l.get(next_pos.0))
                    .unwrap()
                {
                    Field::Empty => {
                        current_pos = next_pos;
                        continue 'current_attempt;
                    }
                    Field::Blocked => {
                        current_direction = current_direction.next();
                        continue 'current_attempt;
                    }
                    Field::Guard(_) => unreachable!(),
                }
            }
            !is_valid
        })
        .count()
}

fn parse_grid(
    input: &str,
    initial_direction: &mut Direction,
    initial_position: &mut (usize, usize),
) -> Vec<Vec<Field>> {
    input
        .lines()
        .enumerate()
        .map(|(row_index, line)| {
            line.chars()
                .enumerate()
                .map(|(x, c)| match parse_field(c) {
                    Field::Empty => Field::Empty,
                    Field::Blocked => Field::Blocked,
                    Field::Guard(direction) => {
                        *initial_direction = direction;
                        *initial_position = (x, row_index);
                        Field::Empty
                    }
                })
                .collect::<Vec<_>>()
        })
        .collect()
}

#[cfg(test)]
mod tests {
    use std::fs::read_to_string;

    use tracing_test::traced_test;

    use super::{pt1, pt2};

    #[traced_test]
    #[test]
    pub fn test_pt1() {
        let input =
            read_to_string("./inputs/day06/example.txt").expect("Missing example.txt for day06");
        assert_eq!(pt1(&input), 41)
    }
    #[test]
    pub fn test_pt2() {
        let input =
            read_to_string("./inputs/day06/example.txt").expect("Missing example.txt for day06");

        assert_eq!(pt2(&input), 6)
    }
}
